#ifndef __GEOMETREE_H
#define __GEOMETREE_H

#include <yafray/core/extern/bound.h>

__BEGIN_YAFRAY

// PURE BSP
//

class YAFRAYCORE_EXPORT plane_t
{
	public:
		static const int XAXIS=0;
		static const int YAXIS=1;
		static const int ZAXIS=2;
		plane_t() {};
		plane_t(const bound_t &b,PFLOAT cut,int ax):zcut(cut),axis(ax) {};
		void set(const bound_t &b,PFLOAT cut,int ax)
		{
			axis=ax;
			zcut=cut;
		};
		bool cross(const ray_t &ray,PFLOAT minD,PFLOAT maxD,PFLOAT &where,int &side)const
		{
			PFLOAT fz,rz;
			switch(axis)
			{
				case XAXIS:fz=ray.from.x;rz=ray.dir.x;break;
				case YAXIS:fz=ray.from.y;rz=ray.dir.y;break;
				default:fz=ray.from.z;rz=ray.dir.z;break;
			}
			side=(fz<zcut) ? 0 : 1;
			if(rz==0) return false;
			PFLOAT D=(zcut-fz)/rz;
			if(D<0) return false;
			if(D<minD) {side=(~side)&1;return false;}
			if(D>maxD) return false;
			where=D;
			return true;
		};
		PFLOAT zcut;
		int axis;
};

inline bool planeCrossInv(PFLOAT zcut,int axis,const ray_t &invray,
		PFLOAT minD,PFLOAT maxD,PFLOAT &where,int &side)
{
	PFLOAT fz,irz;
	switch(axis)
	{
		case plane_t::XAXIS:fz=invray.from.x;irz=invray.dir.x;break;
		case plane_t::YAXIS:fz=invray.from.y;irz=invray.dir.y;break;
		default:fz=invray.from.z;irz=invray.dir.z;break;
	}
	if(fz == zcut)
		side=(irz>0) ? 0 : 1;
	else
		side=(fz<zcut) ? 0 : 1;
	if(irz==0) return false;
	PFLOAT D=(zcut-fz)*irz;
	if(D<0) return false;
	if(D<minD) {side=(~side)&1;return false;}
	if(D>maxD) return false;
	where=D;
	return true;
}

template<class T>
class pureBspTree_t
{
	public:
		pureBspTree_t(T *element):
			left(NULL),right(NULL),leaf(element) {};
		
		pureBspTree_t(pureBspTree_t<T> *l,pureBspTree_t<T> *r,const plane_t &p):
			plane(p),left(l),right(r),leaf(NULL) {};

		~pureBspTree_t() 
		{
			if(left!=NULL) delete _td(left);
			if(right!=NULL) delete _td(right); 
			if(isLeaf()) delete _td(leaf);
		};

		bool isLeaf()const {return leaf!=NULL;};
		PFLOAT getPlane()const {return plane.zcut;};
		int getAxis()const {return plane.axis;};
		const pureBspTree_t<T> *goLeft()const {return left;};
		const pureBspTree_t<T> *goRight()const {return right;};
		const T *getElement()const {return leaf;};
	protected:
		plane_t plane;
		pureBspTree_t<T> *left,*right;
		T *leaf;
};

template<class T>
class pureBspIterator_t
{
	public:
		pureBspIterator_t(const pureBspTree_t<T> *root,/*PFLOAT m,*/const bound_t &bound,
				const segment_t &s,bool sh=false);
		
		void operator ++ ();
		void operator ++ (int) {++(*this);};
		bool operator ! () {return current!=NULL;};
		const T * operator * () {return current;};
		void limit(PFLOAT Z) {if(Z<seg.dist) seg.dist=Z;};
	protected:
		void down(const pureBspTree_t<T> *node,PFLOAT minimun,PFLOAT maximun);
		
		struct state_t
		{
			state_t(const pureBspTree_t<T> *n,PFLOAT e,PFLOAT l):node(n),enter(e),leave(l) {};
			const pureBspTree_t<T> *node; // branch left to go through
			PFLOAT enter;
			PFLOAT leave;
		};
		std::vector<state_t> stack;
		const T *current;
		segment_t seg;
		ray_t invray;
		//bool shadow;
};


template<class T>
pureBspIterator_t<T>::pureBspIterator_t(const pureBspTree_t<T> *root,/*PFLOAT m,*/const bound_t &bound,
																		const segment_t &s,bool sh):
seg(s),invray(s.from,s.dir)//,shadow(sh)
{
	if(root==NULL)
		current=NULL;
	else
	{
		PFLOAT min,max;
		if(!bound.cross(seg,min,max))
			current=NULL;
		else
		{
			if(invray.dir.x!=0) invray.dir.x=1.0/invray.dir.x; else invray.dir.x=0;
			if(invray.dir.y!=0) invray.dir.y=1.0/invray.dir.y; else invray.dir.y=0;
			if(invray.dir.z!=0) invray.dir.z=1.0/invray.dir.z; else invray.dir.z=0;
			seg.dist=(seg.dist<max) ? seg.dist : max;
			stack.reserve(16);
			down(root,min,seg.dist);
		}
	}
}

template<class T>
inline void pureBspIterator_t<T>::down(const pureBspTree_t<T> *node,PFLOAT minimun,PFLOAT maximun)
{
	PFLOAT limit=0;
	int side;
	bool cross;
	while(!node->isLeaf())
	{
		cross=planeCrossInv(node->getPlane(),node->getAxis(),
												invray,minimun,maximun,limit,side);
		if(side==0) //left
		{
			if(cross)
				stack.push_back(state_t(node->goRight(),limit,maximun));
			node=node->goLeft();
		}
		else
		{
			if(cross)
				stack.push_back(state_t(node->goLeft(),limit,maximun));
			node=node->goRight();
		}
		if(cross) maximun=limit;
	}
	current=node->getElement();
}

template<class T>
void pureBspIterator_t<T>::operator ++ ()
{
	current=NULL;
	if(stack.size() && (stack.back().enter<seg.dist))
	{
		const pureBspTree_t<T> *node=stack.back().node;
		PFLOAT enter=stack.back().enter;
		PFLOAT leave=stack.back().leave;
		if(seg.dist<leave) leave=seg.dist;
		stack.pop_back();
		down(node,enter,leave);
	}
}
__END_YAFRAY
#endif
