/****************************************************************************
 *
 * 			bound.cc: Bound and tree implementation for raytracing acceleration
 *      This is part of the yafray package
 *      Copyright (C) 2002  Alejandro Conty Estvez
 *
 *      This library is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU Lesser General Public
 *      License as published by the Free Software Foundation; either
 *      version 2.1 of the License, or (at your option) any later version.
 *
 *      This library is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *      Lesser General Public License for more details.
 *
 *      You should have received a copy of the GNU Lesser General Public
 *      License along with this library; if not, write to the Free Software
 *      Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *      
 */

#include <yafray/core/extern/bound.h>

__BEGIN_YAFRAY

#define MIN(a,b) ( ((a)<(b)) ? (a) : (b))
#define MAX(a,b) ( ((a)>(b)) ? (a) : (b))

bound_t::bound_t(const bound_t &s)
{
	a=s.a;
	g=s.g;
	null=s.null;
}

bound_t::bound_t(const bound_t &r,const bound_t &l)
{
	PFLOAT minx=MIN(r.a.x,l.a.x);
	PFLOAT miny=MIN(r.a.y,l.a.y);
	PFLOAT minz=MIN(r.a.z,l.a.z);
	PFLOAT maxx=MAX(r.g.x,l.g.x);
	PFLOAT maxy=MAX(r.g.y,l.g.y);
	PFLOAT maxz=MAX(r.g.z,l.g.z);
	a.set(minx,miny,minz);
	g.set(maxx,maxy,maxz);
}

GFLOAT bound_t::vol() const
{
	GFLOAT ret=(g.y-a.y)*(g.x-a.x)*(g.z-a.z);
	//if( ret<0 ) cout<<"warning usorted bounding points\n";
	return ret;
}

bool bound_t::cross(const segment_t &seg,PFLOAT &where)const
{
		const point3d_t &a0=a,&a1=g;
		vector3d_t p;
		p=seg.from-a0;

    PFLOAT lmin=-1, lmax=-1,tmp1,tmp2;
		if(seg.dir.x!=0.0)
		{
	    tmp1 =               -p.x/seg.dir.x;
	    tmp2 = ((a1.x-a0.x)-p.x)/seg.dir.x;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    lmin = tmp1;
	    lmax = tmp2;	// si  < 0, podemos devolver false
			if((lmax<0) || (lmin>seg.dist)) return false;
		}
		if(seg.dir.y!=0.0)
		{
	    tmp1 = -p.y/seg.dir.y;
	    tmp2 = ((a1.y-a0.y)-p.y)/seg.dir.y;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    if (tmp1 > lmin)
	        lmin = tmp1;
	    //if (tmp2 < lmax)
	    if ((tmp2 < lmax) || (lmax<0))
	        lmax = tmp2;	// si  < 0, podemos devolver false
			if((lmax<0) || (lmin>seg.dist)) return false;
		}
		if(seg.dir.z!=0.0)
		{
	    tmp1 = -p.z/seg.dir.z;
	    tmp2 = ((a1.z-a0.z)-p.z)/seg.dir.z;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    if (tmp1 > lmin)
	        lmin = tmp1;
	    if ((tmp2 < lmax) || (lmax<0))
	        lmax = tmp2;	
		} 
		if((lmin <= lmax) && (lmax >= 0) && (lmin<=seg.dist))
		{
			where=(lmin>0) ? lmin : 0;
			return true;
		}
		else return false;
}

bool bound_t::cross(const segment_t &seg,PFLOAT &enter,PFLOAT &leave)const
{
		const point3d_t &a0=a,&a1=g;
		vector3d_t p;
		p=seg.from-a0;

    PFLOAT lmin=-1, lmax=-1,tmp1,tmp2;
		if(seg.dir.x!=0.0)
		{
	    tmp1 =               -p.x/seg.dir.x;
	    tmp2 = ((a1.x-a0.x)-p.x)/seg.dir.x;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    lmin = tmp1;
	    lmax = tmp2;	// si  < 0, podemos devolver false
			if((lmax<0) || (lmin>seg.dist)) return false;
		}
		if(seg.dir.y!=0.0)
		{
	    tmp1 = -p.y/seg.dir.y;
	    tmp2 = ((a1.y-a0.y)-p.y)/seg.dir.y;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    if (tmp1 > lmin)
	        lmin = tmp1;
	    //if (tmp2 < lmax)
	    if ((tmp2 < lmax) || (lmax<0))
	        lmax = tmp2;	// si  < 0, podemos devolver false
			if((lmax<0) || (lmin>seg.dist)) return false;
		}
		if(seg.dir.z!=0.0)
		{
	    tmp1 = -p.z/seg.dir.z;
	    tmp2 = ((a1.z-a0.z)-p.z)/seg.dir.z;
	    if (tmp1 > tmp2)
	       std::swap(tmp1, tmp2);
	    if (tmp1 > lmin)
	        lmin = tmp1;
	    if ((tmp2 < lmax) || (lmax<0))
	        lmax = tmp2;	
		} 
		if((lmin <= lmax) && (lmax >= 0) && (lmin<=seg.dist))
		{
			enter=(lmin>0) ? lmin : 0;
			leave=lmax;
			return true;
		}
		else return false;
}

__END_YAFRAY
