#include <yafray/core/intern/path.h>
#include <yafray/core/extern/tools.h>

using namespace std;

__BEGIN_YAFRAY


static vector3d_t sampleLambertian(const vector3d_t &N, const vector3d_t &U, const vector3d_t &V)
{
	PFLOAT z1 = ourRandom(), z2 = ourRandom()*2.0*M_PI;
	
	return (U*cos(z2) + V*sin(z2))*z1 + N * sqrt(1.0-z1*z1);
}

static vector3d_t chooseDirection(const surfacePoint_t &sp, const vector3d_t &N,
																	const vector3d_t &entrance, CFLOAT brdf, 
																	color_t &energy, CFLOAT &prob, CFLOAT &rprob, bool &isdiff)
{
	CFLOAT diff = sp.light.diffuse.energy();
	CFLOAT ref =  sp.light.reflected.energy();
	CFLOAT trans =sp.light.transmitted.energy();
	CFLOAT selection = ourRandom();
	vector3d_t dir;
	
	if(selection< diff)
	{
		dir = sampleLambertian(N, sp.ext_geometry.NU, sp.ext_geometry.NV);
		CFLOAT ocos = dir * N;
		CFLOAT icos = entrance * N;
		energy *= sp.light.diffuse * ocos * brdf;
		prob = diff * ocos / M_PI;
		rprob = diff * icos / M_PI;
		isdiff = true;
	}
	else if( selection < (diff+ref) )
	{
		dir = randomVectorCone(reflect(N, entrance), 
														cos(sp.light.ref_cone), ourRandom(), ourRandom());
		energy *= sp.light.reflected;
		rprob = prob = ref ; // not true, but it is compensated by the same error in the BRDF = 1
		isdiff = false;
	}
	else if( selection < (diff+ref+trans) )
	{
		dir = randomVectorCone(refract(sp.geometry.N, entrance, 
										sp.light.IOR), cos(sp.light.ref_cone), ourRandom(), ourRandom());
		energy *= sp.light.transmitted;
		rprob = prob = trans; // not true, but it is compensated by the same error in the BRDF = 1
		isdiff = false;
	}
	else prob = 0.0; // stop the walk
	return dir;
}

static void temporaryPaint(surfacePoint_t &sp)
{
	sp.light.diffuse.set(0.9, 0.9, 0.9);
	sp.light.reflected.set(0,0,0);
	sp.light.transmitted.set(0,0,0);
	sp.light.ref_cone = 0;
	sp.light.trans_cone = 0;
	sp.light.IOR = 1.0;
}

void eyePath_t::createPath(const scene_t &scene, const ray_t & ray, eyePath_t &path, 
	int maxdepth)
{
	surfacePoint_t sp;
	path.reset();
	energy_t energy;
	energy.segment = ray;
	energy.color.set(1,1,1);
	
	if(!scene.shoot(ray, sp)) return;

	temporaryPaint(sp);
	vector3d_t N = FACE_FORWARD(sp.geometry.Ng, sp.geometry.N, -energy.segment.dir);
	CFLOAT icos = -( N * energy.segment.dir);
	CFLOAT dist = sp.geometry.Z;
	if(icos < 0.0) icos = 0.0;
	energy.color *= icos / (dist * dist); // variable change in the integral
	path.add(energy.color,
					 sp.light.diffuse, sp.geometry.P, N, 
					 (icos / (dist * dist)), // reverse hit prob
					 1.0 / M_PI, sp.basic.origin ); //BRDF
	
	for(int i=0;i<maxdepth;++i)
	{
		eyePath_t::node_t & last = path.path().back();
		CFLOAT prob = 1.0;
		CFLOAT rprob = 1.0;
		energy.segment.from = sp.geometry.P;
		energy.segment.dir = chooseDirection(sp, N, -energy.segment.dir, 
																				 1.0 / M_PI, energy.color, prob, rprob, last.diffuse_jump);

		if(prob == 0.0) break; // ABSORTION
		if(!scene.shoot(energy.segment, sp, sp.basic.origin)) break;
		
		temporaryPaint(sp);
		N = FACE_FORWARD(sp.geometry.Ng, sp.geometry.N, -energy.segment.dir);
		CFLOAT icos = -(energy.segment.dir * N);
		CFLOAT ocos = fabs(energy.segment.dir * last.N);
		CFLOAT dist = sp.geometry.Z;
		if(icos < 0.0) icos = 0.0;

		last.rtrans_prob = prob;
		last.trans_prob = rprob;
		last.hit_prob = ocos / (dist * dist);

		energy.color *= icos / (dist * dist); // variable change in the integral
		path.add(energy.color,
						 sp.light.diffuse, sp.geometry.P, N, 
						 (icos / (dist * dist)), // reverse hit prob
						 1.0 / M_PI, sp.basic.origin ); //BRDF

		// In case walk ends, last node is diffuse, so transmission probability is :				 
		path.path().back().trans_prob = sp.light.diffuse.energy() * icos / M_PI;
	}
}

void lightPath_t::createPath(const scene_t &scene, light_t *L, 
														lightPath_t &path, int maxdepth)
{

	surfacePoint_t sp;
	path.reset();
	energy_t energy;
	L->getSample(energy);
	
	path.add(energy.color, color_t(1,1,1), energy.segment.from, vector3d_t(0,0,1) , 
					1.0 , 1 / (4.0*M_PI), NULL);

	energy.color *= 1.0 / (4.0*M_PI);
	vector3d_t N;
	if( (maxdepth > 0) && scene.shoot(energy.segment, sp, NULL) )
	{
		lightPath_t::node_t & last = path.path().back();
		temporaryPaint(sp);
		N = FACE_FORWARD(sp.geometry.Ng, sp.geometry.N, -energy.segment.dir);
		CFLOAT icos = -( N * energy.segment.dir);
		CFLOAT dist = sp.geometry.Z;
		if(icos < 0.0) icos = 0.0;
		energy.color *= icos / (dist * dist); // variable change in the integral
		
		last.trans_prob = 1.0 / (4.0 * M_PI); // hit probs for the light are 0
	
		path.add(energy.color ,
						 sp.light.diffuse, sp.geometry.P, N, 
						 (icos / (dist * dist)), // hit prob
						 1.0 / M_PI, sp.basic.origin ); //BRDF
	}
	
	for(int i=1;i<maxdepth;++i)
	{
		lightPath_t::node_t & last = path.path().back();
		CFLOAT prob = 1.0;
		CFLOAT rprob = 1.0;
		
		energy.segment.from = sp.geometry.P;
		energy.segment.dir = chooseDirection(sp, N, -energy.segment.dir, 
																				 1.0 / M_PI, energy.color, prob, rprob, last.diffuse_jump);
		
		if(prob == 0.0) break; // ABSORTION
		if(!scene.shoot(energy.segment, sp, sp.basic.origin)) break;
		temporaryPaint(sp);
		N = FACE_FORWARD(sp.geometry.Ng, sp.geometry.N, -energy.segment.dir);
		CFLOAT icos = -(energy.segment.dir * N);
		CFLOAT ocos = fabs(energy.segment.dir * last.N);
		CFLOAT dist = sp.geometry.Z;
		if(icos < 0.0) icos = 0.0;

		last.trans_prob = prob;
		last.rtrans_prob = rprob;
		last.rhit_prob = ocos / (dist * dist);
		
		energy.color *= icos / (dist * dist); // variable change in the integral
		path.add(energy.color ,
						 sp.light.diffuse, sp.geometry.P, N, 
						 (icos / (dist * dist)), // hit prob
						 1.0 / M_PI, sp.basic.origin ); //BRDF

		// In case walk ends, last node is diffuse, so transmission probability is :				 
		path.path().back().rtrans_prob = sp.light.diffuse.energy() * icos / M_PI;
	}
}


__END_YAFRAY
