#include <yafray/core/extern/scene.h>
#include <yafray/core/extern/tools.h>
#include <yafray/core/intern/meshtree.h>
#include <yafray/core/intern/pointlight.h>

#include <iostream>
#include <limits>

using namespace std;

__BEGIN_YAFRAY

scene_t::scene_t()
{
	geometry.tree=NULL;
	state_stack.push_front(state::NONE);
	triangle_t::setContext(geometry.points, geometry.orcos, geometry.colors);
}

scene_t::~scene_t()
{
	if(getState()!=state::NONE) cout<<"Destroyed scene in unstable state"<<endl;
	if(geometry.tree!=NULL) delete _td((tnode_t *)geometry.tree);
	v_foreach(light_t *, i, lights)
		delete _td(*i);
}

void scene_t::calculateBound()
{
	PFLOAT maxX=-numeric_limits<PFLOAT>::infinity(),minX=numeric_limits<PFLOAT>::infinity();
	PFLOAT maxY=-numeric_limits<PFLOAT>::infinity(),minY=numeric_limits<PFLOAT>::infinity();
	PFLOAT maxZ=-numeric_limits<PFLOAT>::infinity(),minZ=numeric_limits<PFLOAT>::infinity();

	v_foreach(point3d_t, i, geometry.points)
	{
		if(i->x<minX) minX=i->x; if(i->x>maxX) maxX=i->x;
		if(i->y<minY) minY=i->y; if(i->y>maxY) maxY=i->y;
		if(i->z<minZ) minZ=i->z; if(i->z>maxZ) maxZ=i->z;
	}
	geometry.bound.set(point3d_t(minX,minY,minZ), point3d_t(maxX,maxY,maxZ));
}

void scene_t::startGeometry()
{
	if(geometry.tree!=NULL) delete _td((tnode_t *)geometry.tree);
	geometry.tree=NULL;
	geometry.objects.clear();
	geometry.triangles.clear();
	geometry.colors.clear();
	geometry.orcos.clear();
	geometry.points.clear();
	load.matrix_stack.clear();
	load.matrix_stack.push_front(matrix4x4_t(1.0));
	state_stack.push_front(state::GEOMETRY);
}

void scene_t::startObject()
{
	load.object = geometry.objects.size();
	load.tribegin  = geometry.triangles.size();
	load.pbegin  = geometry.points.size();
	load.obegin  = geometry.orcos.size();
	load.cbegin  = geometry.colors.size();
	load.lasttri = NULL;
}

int scene_t::addVertex(const point3d_t &p)
{
	int index=geometry.points.size(); //-load.pbegin;
	point3d_t trans=load.matrix_stack.front()*p;
	geometry.points.push_back(trans);
	return index;
}

int scene_t::addOrco(const point3d_t &p)
{
	int index=geometry.orcos.size();//-load.obegin;
	geometry.orcos.push_back(p);
	return index;
}

int scene_t::addColor(const color_t &c)
{
	int index=geometry.colors.size();//-load.cbegin;
	geometry.colors.push_back(c);
	return index;
}
		
void scene_t::addTriangle(int pa,int pb,int pc,
										 const vector3d_t &na,const vector3d_t &nb,const vector3d_t &nc)
{
	//geometry.triangles.push_back(triangle_t(pa+load.pbegin,pb+load.pbegin,pc+load.pbegin));
	geometry.triangles.push_back(triangle_t(pa, pb, pc));
	load.lasttri=&(geometry.triangles.back());
	matrix4x4_t &M=load.matrix_stack.front();
	load.lasttri->setNormals(M*na,M*nb,M*nc);
}

void scene_t::setOrco(int oa,int ob,int oc)
{
	//load.lasttri->setOrco(oa+load.obegin,ob+load.obegin,oc+load.obegin);
	load.lasttri->setOrco(oa, ob, oc);
}

void scene_t::setUV(GFLOAT uv[3][2])
{
	load.lasttri->setUV(uv[0][0],uv[0][1],uv[1][0],uv[1][1],uv[2][0],uv[2][1]);
}

void scene_t::setUV(GFLOAT a_u,GFLOAT a_v,GFLOAT b_u,GFLOAT b_v,GFLOAT c_u,GFLOAT c_v)
{
	load.lasttri->setUV(a_u, a_v, b_u, b_v, c_u, c_v);
}
	
void scene_t::setColor(int ca,int cb,int cc)
{
	//load.lasttri->setVCOL(ca+load.cbegin,cb+load.cbegin,cc+load.cbegin);
	load.lasttri->setVCOL(ca, cb, cc);
}
		
void scene_t::endObject()
{
	geometry.objects.push_back(pair<int,int>(load.tribegin, geometry.triangles.size()));
}

void scene_t::endGeometry()
{
	vector<triangle_t *> *tri=_tn(new vector<triangle_t *>);
	tri->reserve(geometry.triangles.size());
	
	v_foreach(triangle_t, i, geometry.triangles)
		(*tri).push_back(&(*i));
	calculateBound();
	geometry.tree=buildTriangleTree(tri,geometry.bound,4);
	state_stack.pop_front();
}

configurable_t * scene_t::addLight(const char *type)
{
// Don't fear, this will be changed
	string T = type;
	light_t *L = NULL;
	if(T == "point") L = _tn(new pointLight_t());
	if(L != NULL) lights.push_back(L);
	return L;
}


__END_YAFRAY
