/****************************************************************************
 *
 * 			triangle.cc: Face representation and manipulation implementation
 *      This is part of the yafray package
 *      Copyright (C) 2002 Alejandro Conty Estevez
 *
 *      This library is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU Lesser General Public
 *      License as published by the Free Software Foundation; either
 *      version 2.1 of the License, or (at your option) any later version.
 *
 *      This library is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *      Lesser General Public License for more details.
 *
 *      You should have received a copy of the GNU Lesser General Public
 *      License along with this library; if not, write to the Free Software
 *      Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <yafray/core/extern/triangle.h>

using namespace std;

__BEGIN_YAFRAY

const vector<point3d_t> * triangle_t::points=NULL;
const vector<point3d_t> * triangle_t::orcos=NULL;
const vector<color_t> * triangle_t::colors=NULL;

/*
void triangle_t::check()const
{
	if( (pa<0) || (pa>=points->size()) ||
			(pb<0) || (pb>=points->size()) ||
			(pc<0) || (pc>=points->size()) )
			cout<<"triangulo corrupto "<<pa<<" "<<pb<<" "<<pc<<endl;
}
*/
triangle_t::triangle_t(int va,int vb,int vc)
{
	pa=va;
	pb=vb;
	pc=vc;
	const point3d_t &a=(*points)[pa],&b=(*points)[pb],&c=(*points)[pc];
	normal=(b-a)^(c-a);
	normal.normalize();
	//uv=NULL;
	hasuv = false;
	has_vcol = false;
	has_orco = false;
}

void triangle_t::recNormal()
{
	const point3d_t &a=(*points)[pa],&b=(*points)[pb],&c=(*points)[pc];
	normal=(b-a)^(c-a);
	normal.normalize();
}

triangle_t::triangle_t():normal(0,0,0)
{
	hasuv = false;
	has_vcol = false;
	has_orco=false;
}

void triangle_t::setVertices(int va,int vb,int vc)
{
	pa=va;
	pb=vb;
	pc=vc;
	const point3d_t &a=(*points)[pa],&b=(*points)[pb],&c=(*points)[pc];
	normal=(b-a)^(c-a);
	normal.normalize();
}

void triangle_t::setUV(GFLOAT a_u,GFLOAT a_v,GFLOAT b_u,GFLOAT b_v,GFLOAT c_u,GFLOAT c_v)
{
	uv[0][0]=a_u;
	uv[0][1]=a_v;
	uv[1][0]=b_u;
	uv[1][1]=b_v;
	uv[2][0]=c_u;
	uv[2][1]=c_v;
	hasuv=true;
}

void triangle_t::setVCOL(int a,int b,int c)
{
	ca=a;
	cb=b;
	cc=c;
	has_vcol=true;
}

void triangle_t::setOrco(int a,int b,int c)
{
	oa=a;
	ob=b;
	oc=c;
	has_orco=true;
}


static bool getInterpolation(const point3d_t &a,const point3d_t &b,const point3d_t &c
		,const point3d_t &h,GFLOAT &fa,GFLOAT &fb,GFLOAT &fc)
{
	vector3d_t va=a-h;
	vector3d_t vb=b-h;
	vector3d_t vc=c-h;
	vector3d_t v1(va.x,vb.x,vc.x);
	vector3d_t v2(va.y,vb.y,vc.y);
	vector3d_t v3(va.z,vb.z,vc.z);
	va=v1^v2;  va.abs();
	vb=v2^v3;  vb.abs();
	vc=v1^v3;  vc.abs();
	PFLOAT smod1=va.x+va.y+va.z;
	PFLOAT smod2=vb.x+vb.y+vb.z;
	PFLOAT smod3=vc.x+vc.y+vc.z;
	if( (smod1>smod2) && (smod1>smod3)) { v1=va; }
	else if( (smod2>smod1) && (smod2>smod3)) { smod1=smod2;  v1=vb; }
	else { smod1=smod3;  v1=vc; }
	if(smod1==0.0)
		return false;
	v1/=smod1;
	fa=v1.x;  fb=v1.y;  fc=v1.z;
	return true;
}

surfacePoint_t  triangle_t::getSurface(point3d_t &h,PFLOAT d)const
{
	surfacePoint_t sp;
		
	sp.basic.object = object;
	//sp.basic.shader;
	//sp.basic.screenpos;
	sp.basic.origin = (void *)this;
	
	const point3d_t &a=(*points)[pa],&b=(*points)[pb],&c=(*points)[pc];
	GFLOAT fa,fb,fc;
	if (!getInterpolation(a, b, c, h, fa, fb, fc))
		return sp;
		
	vector3d_t nn = na*fa+nb*fb+nc*fc;
	nn.normalize();

	sp.geometry.available = true;
	sp.geometry.Z = d;
	sp.geometry.N = nn;
	sp.geometry.Ng = normal;
	sp.geometry.P = h;

	point3d_t orcoP;
	if(has_orco) orcoP=(*orcos)[oa]*fa+(*orcos)[ob]*fb+(*orcos)[oc]*fc; else orcoP=h;
	
	GFLOAT u=0, v=0;
	if (hasuv)
	{
		u=uv[0][0]*fa+uv[1][0]*fb+uv[2][0]*fc;
		v=uv[0][1]*fa+uv[1][1]*fb+uv[2][1]*fc;
	}
	color_t vcolor(0.0);
	if (has_vcol)
	{
		const color_t &a=(*colors)[ca],&b=(*colors)[cb],&c=(*colors)[cc];
		vcolor=a*fa+b*fb+c*fc;
	}
	sp.geometry.orcoP = orcoP;
	sp.geometry.vertex_col = vcolor;
	sp.geometry.U = u;
	sp.geometry.V = v;
	sp.geometry.hasUV = hasuv; 
	sp.geometry.hasOrco = has_orco;
	sp.geometry.hasVertexCol = has_vcol;
	
	sp.ext_geometry.available = true;
	createCS(sp.geometry.N, sp.ext_geometry.NU, sp.ext_geometry.NV);
	if(hasuv)
	{
		vector3d_t vb=b-a;
		vector3d_t vc=c-a;
		GFLOAT lenb=vb.length();

		GFLOAT dub=(uv[1][0]-uv[0][0])/lenb;
		GFLOAT dvb=(uv[1][1]-uv[0][1])/lenb;
		vb/=lenb;
		GFLOAT nuc=uv[2][0],nvc=uv[2][1],proj=vc*vb;

		vc=vc-vb*proj;
		nuc-=proj*dub;
		nvc-=proj*dvb;

		GFLOAT lenc=vc.length();
		GFLOAT duc=(nuc-uv[0][0])/lenc;
		GFLOAT dvc=(nvc-uv[0][1])/lenc;
		vc/=lenc;

		GFLOAT projC=sp.ext_geometry.NU*vc,projB=sp.ext_geometry.NU*vb;
		GFLOAT dudu=projC*duc+projB*dub;
		GFLOAT dvdu=projC*dvc+projB*dvb;

		projC=sp.ext_geometry.NV*vc,projB=sp.ext_geometry.NV*vb;
		GFLOAT dudv=projC*duc+projB*dub;
		GFLOAT dvdv=projC*dvc+projB*dvb;

		sp.ext_geometry.dudNU = dudu;
		sp.ext_geometry.dudNV = dudv;
		sp.ext_geometry.dvdNU = dvdu;
		sp.ext_geometry.dvdNV = dvdv;
	}

	sp.light.available = false;

	return sp;
}

__END_YAFRAY
